
import os
import xml

import logging
import xml.etree.ElementTree as ET

from mypy_gpt.common import setup_logger, generate_diff
from mypy_gpt.guide import Guidance
from mypy_gpt.linter import MyPyLinter

logger=logging.getLogger('mypygpt')
logger.propagate=False

DEFAULT_MODEL = "gpt-3.5-turbo-16k"

MYPYARGS = ['--disallow-untyped-defs']

older_path = r"c:\gitproj\Auto-GPT"
DEFAULT_TOKENS =3600
DEFAULT_TOKENS_PER_FIX =400
DEFAULT_TEMP_PER_FIX=0.7
DEFAULT_TEMP = 0.2
import argparse
from typing import List, Dict, Any, Optional, Iterator


class MyPyGpt:
    def __init__(self,args: argparse.Namespace):
        #move all the attributes of args to local variables
        self.args = args
        self.file = args.file
        self.original_content = open(self.file, 'rt').read()

        self.debug = args.debug
        self.linter = MyPyLinter(args.debug)

    def get_new_content(self,err_res: Dict[str, Any]) -> Optional[str]:
        fix_guide= Guidance.guide_for_fixes(self.args)
        fix_res=fix_guide(filename=self.file, file=self.original_content, fixes=err_res['fix'])
        if not 'fixedfile' in fix_res:
            logger.error('no fixed file')
            return None
        fixed = fix_res['fixedfile']
        logger.debug(f'fixed file: {fixed}')
        #bb = json.loads(fix_res["fixedfile"])['pythonfile']
        for k in range(3):

            if k==1:
                try:
                    logger.debug("trying to fix bad formatting")
                    new_content=fixed[fixed.index('<pythonfile>'):]
                    new_content=new_content[:new_content.rindex('</pythonfile>')+len('</pythonfile>')]
                except Exception as e :
                    logger.error(e)
                    logger.error(f"bad formatting manual extraction {k}")
                    continue
            elif k==0:
                try:
                    new_content= fixed[fixed.index('```xml') + 6:]
                    new_content=new_content[:new_content.rindex('```')]
                except:
                    logger.debug("no ```xml found ")
                    if len(fixed)>0.5*len(self.original_content):
                        new_content=fixed
                    else:
                        return None
            elif k==2:
                try:
                    new_content= fixed[fixed.index('```python') + 9:] #it wasn't supposed to be like that
                    new_content=new_content[:new_content.rindex('```')]
                except:
                    logger.debug("no ```python found ")

            try:
                new_content=ET.fromstring(new_content).text #remove the pythonfile element
                return new_content

            except xml.etree.ElementTree.ParseError:
                logger.debug(f"bad formatting {k}")
                logger.debug(new_content)
                if new_content.startswith('<pythonfile>') and new_content.endswith('</pythonfile>') and len(new_content) > 0.8 * len(self.original_content):
                    new_content= new_content[len('<pythonfile>'):(-1)*len('</pythonfile>')] #it is probably too idiot to extract it.
                    return new_content
                if k == 2 and len(new_content) > 0.8 * len(self.original_content):
                    logger.warn("will try anyway")
                    return new_content

        logger.debug("got to the end")
        return None



    def get_issues_string(self,issues: List[Dict[str, Any]]) -> Iterator[str]:
        for issue in issues:
            ln=int(issue['Line Number'])
            lines=self.original_content.split('\n')
            line_range= '\n'.join( lines[ max(ln-1-1,0) :min(ln-1+1,len(lines))])

            issue[f"lines {ln-1} to {ln} in the file"]='\n'+line_range

            st='\n'.join(f"{k}: {v}" for k,v in issue.items())

            logger.debug(st)
            yield st

    def main(self) -> None:
        setup_logger(logger,self.debug)

        if 'OPEN_AI_KEY' not in os.environ:
            logger.error('OPEN_AI_KEY not set')
            return
        Guidance.set_key(self.args)
        logger.info("mypy output:")

        errors = self.linter.get_issues(self.args)
        logger.debug(errors)
        if len(errors)==0:
            logger.info('no errors')
            return

        self.try_to_solve_issues(errors)

    def try_to_solve_issues(self,errors):
        logger.info("trying to solve issues") 
        err_res = self.get_fixes(list(self.get_issues_string(errors)))
        new_content=self.get_new_content(err_res)
        if new_content is None:
            logger.error('cant continue')
            return


        colored_diff,diff= generate_diff(self.original_content, new_content, self.args.file.replace("\\", '/'))

        if self.args.diff_file:
            open(self.args.diff_file, 'wt').write(diff)

        old_errors=errors
        if not self.args.recheck_policy == 'none':
            errors = self.check_new_file(new_content)

         
        print(diff if self.args.no_color else colored_diff)
        update=False


        if (len(errors) == 0 and self.args.auto_update == 'strict') \
            or (self.args.auto_update == 'permissive' and len(old_errors) > len(errors)):
                update=True

        elif not self.args.dont_ask:
            print("do you want to override the file? (y/n)")
            if input() == 'y':
                update=True

        if update:
            open(self.args.file, 'wt').write(new_content)
            self.original_content = new_content
            if self.args.recheck_policy =='recheckandloop' and len(errors) >0:
                self.try_to_solve_issues(errors)

    def check_new_file(self, new_content: str) -> List[Dict[str, Any]]:
        newfile: str = self.args.file.replace('.py', '.fixed.py')  # must be in the same folder sadly.
        open(newfile, 'wt').write(new_content)
        logger.info('output from mypy after applying the fixes:')
        try:
            return self.linter.get_issues(self.args, override_file=newfile)
        finally:
            if not self.args.store_fixed_file:
                try:
                    os.remove(newfile)
                except:
                    logger.error('could not remove file %s' % newfile)


    def get_fixes(self, errors: List[str]) -> Dict[str, Any]:
        err_guide = Guidance.guide_for_errors(self.args)
        err_res: Dict[str, Any] = err_guide(filename=self.file, file=self.original_content, errors=errors)
        if not self.args.dont_print_fixes:
            logger.info('suggested fixes:')
            logger.info('\n'.join(err_res['fix']))
        return err_res


def main() -> None:
    # Create the argument parser
    parser: argparse.ArgumentParser = argparse.ArgumentParser(description='Run mypy on a Python file and use OpenAI GPT to fix the errors. It temporary generates file.fixed.py file to check for errors. You probably want to provide project so that mypy could resolve dependencies.')
    # Add the arguments
    parser.add_argument('file', help='Python file to run mypy on')
    parser.add_argument('mypy_args', nargs=argparse.REMAINDER, help=f'Additional options for mypy after --. By default, uses {MYPYARGS}')
    parser.add_argument('--mypy-path', default='mypy', help='Path to mypy executable (default: "mypy")')
    parser.add_argument('--error_categories', action='store', help='Type of errors to process')
    parser.add_argument('--max_errors', action='store', type=int, default=10, help='Max number of errors to process per cycle')
    parser.add_argument('-p','--proj-path', default='.', help='Path to project')
    parser.add_argument('-d','--diff-file', action='store', help='Store diff in diff file')
    parser.add_argument('-s','--store-fixed-file', action='store_true', default=False, help='Keeps file.fixed.py')

    parser.add_argument('--dont-ask', action='store_true', default=False,
                        help='Dont ask if to apply to changes. Useful for generting diff')
    parser.add_argument('-m','--model', default=DEFAULT_MODEL, help='Openai model to use')
    parser.add_argument('--max_tokens-per-fix', default=DEFAULT_TOKENS_PER_FIX, help='tokens to use for generating each fix')
    parser.add_argument('--temperature-per-fix', default=DEFAULT_TEMP_PER_FIX, help='temperature to use for fixes')
    parser.add_argument('--max_tokens-for-file', default=DEFAULT_TOKENS, help='tokens to use for file')
    parser.add_argument('--temperature-for-file', default=DEFAULT_TEMP, help='temperature to use for generating the file')
    parser.add_argument('-r','--recheck-policy', choices=['recheck','none','recheckandloop'],  default='recheckandloop',
                        help='Recheck the file for issues before suggesting a fix. require to temporarily save file.fixed.py (has to be in the project). recheckandloop will go for another loop if done fixing and there are still errors.')

    parser.add_argument('--debug', action='store_true', default=False, help='debug log level ')
    parser.add_argument('-a','--auto-update', choices=['permissive','no','strict'], default='no', help='auto update file if no errors (if strict). On permissive policy it updates if the number of errors decreased. ')
    parser.add_argument('-D','--dont-print-fixes', action='store_true', default=False, help='dont print fixes')
    #add no colors option  
    parser.add_argument('-N','--no_color', action='store_true', help='dont print color diff')

    # Parse the arguments
    args: argparse.Namespace = parser.parse_args()
    if len(args.mypy_args) ==0:
        args.mypy_args = MYPYARGS

    MyPyGpt(args).main()


if __name__ == '__main__':
    main()



import os

import guidance


class Guidance:
    @staticmethod
    def set_key(args):
        guidance.llm = guidance.llms.OpenAI(args.model, api_key=os.environ['OPEN_AI_KEY'])


    @staticmethod
    def guide_for_errors(args):

        text = '''
    {{#system~}}
        You are a helpful assistant. You will be given a file and an issue. You need to come up with fixes for the issue, even if it is a minor issue.
        {{~/system}}
    {{~#user}}
    Given this {{filename}}: .
            {{file}}.
    A list of issues will be given
    {{~/user}}
    
    {{#each errors}}
      {{~#user}}
      What is the fix for this issue on {{filename}}?
              {{this}}
       Be sure to inspect the entire relevant function before suggesting a fix. 
        Be short and precise regarding the fix, and refer to the change in code. In your answer, you shouldn't repeat instructions given to you,  and you shouldn't include more than 5 lines of code.
      {{~/user}}
      {{#assistant~}}
        {{gen 'fix' list_append=True temperature=%d max_tokens=%d}}
        {{~/assistant}}
    {{/each~}}''' % (args.temperature_per_fix, args.max_tokens_per_fix)
        return  guidance(text, log=True, caching=False)  # type: ignore

    @staticmethod
    def guide_for_fixes(args):
        text = '''
            {{#system~}}
            You are a helpful assistant. You will be given a list of corrections to do in a file, and will update the file accordingly. 
            Reply only with xml that has the following format:  
            ```xml
            <pythonfile>the updated file content after the corrections are made</pythonfile>
            ```
            {{~/system}}
            {{#user~}}
            This is the file:
            {{file}}
            Those are the fixes
            {{#each fixes}}- {{this}}
                {{/each~}}
            Make sure you apply all the corrections in the resulted file, even if the issues aren't clear.
            {{~/user}}
    
            {{#assistant~}}
            {{gen 'fixedfile' temperature=%d max_tokens=%d}}
            {{~/assistant~}}
        '''
        return guidance(text % (args.temperature_for_file , args.max_tokens_for_file), log=True, caching=False)  # type: ignore




import difflib
import logging
import os
from typing import Tuple, Iterable

def setup_logger(logger: logging.Logger, debug: bool) -> None:
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    log_format = "%(levelname)s - %(message)s"
    formatter = logging.Formatter(log_format)
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.handlers = [ch]

def generate_diff(original_content: str, new_content: str, path: str) -> Tuple[str, str]:
    try:
        from colorama import Fore, init
        init()
    except ImportError:  # fallback so that the imported classes always exist
        class ColorFallback():
            __getattr__ = lambda self, name: ''

        Fore = ColorFallback()

    def color_diff(diff: Iterable[str]) -> Iterable[str]:
        for line in diff:
            if line.startswith('+'):
                yield Fore.GREEN + line + Fore.RESET
            elif line.startswith('-'):
                yield Fore.RED + line + Fore.RESET
            elif line.startswith('^'):
                yield Fore.BLUE + line + Fore.RESET
            elif line.startswith('@@'):
                yield Fore.BLUE + line[:line.rindex('@@')+2] + Fore.RESET
            else:
                yield line

    diffres = list(difflib.unified_diff(original_content.split('\n'), new_content.split('\n'), fromfile=path, tofile=path+"b", lineterm=''))
    return os.linesep.join(color_diff(diffres)), os.linesep.join(diffres)



from setuptools import setup
with open('requirements.txt') as f:
    requirements = f.read().splitlines()
import pkg_resources
import setuptools
import pathlib
desc = 'Python minor issue resolver (mypy) using gpt by openai!'
long_desc = 'Solves mypy errors using guidance and GPT api. It runs mypy on targeted file and then uses gpt to try to fix the issues (espcially good for minor nagging issues). Displays a diff file for the required changes and ask you if you want to apply.'
with pathlib.Path('requirements.txt').open() as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements_txt)
    ]

setup(
    name='mypy-gpt',
    version='1.0.6',
    packages=['mypy_gpt'],
    url='https://github.com/eyalk11/mypy-gpt',
    license=' AGPL-3.0 license',
    author='ekarni',
    author_email='eyalk5@gmail.com',
    description=desc,
    long_description=long_desc,
    install_requires=install_requires
)




import logging
import os
import re
import subprocess
from abc import ABCMeta, abstractmethod
from typing import List, Dict, Any, Optional

import pandas
from mypy_gpt.common import setup_logger

logger = logging.getLogger('linter')
logger.propagate = False


class Linter(metaclass=ABCMeta):
    @abstractmethod
    def run_checker(self, file: str, additional_args: List[str], program_path: str, proj_path: str) -> str:
        pass

    @staticmethod
    @abstractmethod
    def parse_line(line: str) -> Dict[str, Any]:
        pass

    def get_issues(self, args: Any, override_file: Optional[str] = None) -> List[Dict[str, Any]]:
        out = self.run_checker(args.file if override_file is None else override_file, args.mypy_args, args.mypy_path,
                               args.proj_path)
        logger.info(out)
        issues = [self.parse_line(z) for z in out.split('\n')]
        issues = list(filter(lambda x: x, issues))
        if len(issues) == 0:
            return []
        # Here we unite the issues from different lines
        dfo = pandas.DataFrame(issues)
        changed_message_df = dfo.groupby('Line Number')['Message'].agg(lambda x: '\n'.join(x))
        df_first_row = dfo.groupby('Line Number').first()
        df_first_row['Message'].update(changed_message_df)
        df_first_row = df_first_row.reset_index()
        issues = [dict(r[1]) for r in df_first_row.iterrows()]

        if args.max_errors:
            issues = issues[:args.max_errors]
        if args.error_categories:
            issues = [z for z in issues if z['Category'] in args.error_categories.split(',')]
        return issues


class MyPyLinter(Linter):
    def __init__(self, debug: bool):
        setup_logger(logger, debug)  # for now

    def run_checker(self, file: str, additional_args: List[str], program_path: str, proj_path: str) -> str:
        # Construct the mypy command
        command = [program_path] + additional_args + [file]
        logger.debug("Running mypy command: %s", command)
        # Run mypy command and capture the output
        result: CompletedProcess[str] = subprocess.run(command, capture_output=True, text=True, cwd=os.path.abspath(proj_path))
        # Print the output

        # Decode the output from bytes to string
        output = result.stdout

        # 7-bit C1 ANSI sequences
        ansi_escape = re.compile(r'''
                \x1B  # ESC
                (?:   # 7-bit C1 Fe (except CSI)
                    [@-Z\\-_]
                |     # or [ for CSI, followed by a control sequence
                    \[
                    [0-?]*  # Parameter bytes
                    [ -/]*  # Intermediate bytes
                    [@-~]   # Final byte
                )
            ''', re.VERBOSE)
        result: str = ansi_escape.sub('', output)
        # Print the output
        return result

    @staticmethod
    def parse_line(line: str) -> None:
        import re
        # Extracting message, type, and line number using regular expressions
        pattern = r'(.+):(\d+): (\w+): (.+) \[(.*)\]'
        match = re.match(pattern, line)
        if not match:
            pattern = r'(.+):(\d+): (\w+): (.+)()'
            match = re.match(pattern, line)

        if match:
            linenumber = match.group(2)
            error_type = match.group(3)
            message = match.group(4)
            sub_type = match.group(5)
            return {"Line Number": linenumber, "Error Type": error_type, "Message": message, "Category": sub_type}
        else:
            logger.debug(("No match found.", line))
